#include <iostream>
#include <algorithm>

#define COUNTER_TRESHOLD 100
#define HAMMER 0
#define SCREW_DRIVER 1
#define SAW 2
#define TOOLS_COUNT 3
#define THREADS_COUNT 3
#define WORK_ITERATIONS 1000000

using namespace std;

struct argsStruct_t {
    pthread_mutex_t *counterMutex;
    pthread_mutex_t *toolMutex;
    pthread_cond_t *toolCondition;
    int *counter;
    int *tool1;
    string tool1Name;
    int *tool2;
    string tool2Name;
    int threadID;
};

void work() {
    for (int i = 0; i < WORK_ITERATIONS; i++);
}

void borrowTools(int **tools, const int toolsCount, pthread_mutex_t *mutex, pthread_cond_t *condition)
{
    pthread_mutex_lock(mutex);
    while ((accumulate(tools, tools+toolsCount, 1, [](int a, int* b){return min(a, *b);})) < 1)
        pthread_cond_wait(condition, mutex);
    for(int i = 0; i < toolsCount; i++) {
        (*tools[i]) -= 1;
    }
    cout << endl;
    pthread_mutex_unlock(mutex);
}

void returnTools(int **tools, const int toolsCount, pthread_mutex_t *mutex, pthread_cond_t *condition)
{
    pthread_mutex_lock(mutex);
    for(int i = 0; i < toolsCount; i++) {
        (*tools[i]) += 1;
    }
    cout << endl;
    pthread_cond_broadcast(condition);
    pthread_mutex_unlock(mutex);
}

void *guyThread(void *args) {
    argsStruct_t *tool = (argsStruct_t *) args;
    int* toolsArgs[2] = {tool->tool1, tool->tool2};
    while (true) {
        borrowTools(&tool->tool1, 1, tool->toolMutex, tool->toolCondition);
        cout << "Guy " << tool->threadID << " borrowed " << tool->tool1Name << "." << endl;
        work();
        returnTools(&tool->tool1, 1, tool->toolMutex, tool->toolCondition);
        cout << "Guy " << tool->threadID << " returned " << tool->tool1Name << "." << endl;

        borrowTools(toolsArgs, 2, tool->toolMutex, tool->toolCondition);
        cout << "Guy " << tool->threadID << " borrowed " << tool->tool1Name << " and " << tool->tool2Name << "."
             << endl;
        work();
        returnTools(toolsArgs, 2, tool->toolMutex, tool->toolCondition);
        cout << "Guy " << tool->threadID << " returned " << tool->tool1Name << " and " << tool->tool2Name << "."
             << endl;
        if ((*tool->counter) > COUNTER_TRESHOLD)
            break;
        pthread_mutex_lock(tool->counterMutex);
        (*tool->counter)++;
        pthread_mutex_unlock(tool->counterMutex);
    }
    return 0;
}

int main() {

    pthread_mutex_t toolMutex;
    pthread_mutex_t counterMutex;
    pthread_cond_t toolCondition;
    pthread_mutex_init(&toolMutex, NULL);
    pthread_mutex_init(&counterMutex, NULL);
    pthread_cond_init(&toolCondition, NULL);

    int toolsCount[TOOLS_COUNT];
    for (int i = 0; i < TOOLS_COUNT; i++)
        toolsCount[i] = 1;
    int counter = 0;
    argsStruct_t threadTools[THREADS_COUNT];
    threadTools[0] = {&counterMutex, &toolMutex, &toolCondition, &counter, &toolsCount[HAMMER], "hammer",
                      &toolsCount[SCREW_DRIVER], "screw driver", 0};
    threadTools[1] = {&counterMutex, &toolMutex, &toolCondition, &counter, &toolsCount[SCREW_DRIVER], "screw driver",
                      &toolsCount[SAW], "saw", 1};
    threadTools[2] = {&counterMutex, &toolMutex, &toolCondition, &counter, &toolsCount[SAW], "saw",
                      &toolsCount[HAMMER], "hammer", 2};

    pthread_t threadGuy[THREADS_COUNT];
    for (int i = 0; i < THREADS_COUNT; i++)
        pthread_create(&threadGuy[i], NULL, guyThread, (void *) &threadTools[i]);
    for (int i = 0; i < THREADS_COUNT; i++)
        pthread_join(threadGuy[i], NULL);

    pthread_mutex_destroy(&toolMutex);
    pthread_mutex_destroy(&counterMutex);
    pthread_cond_destroy(&toolCondition);
    return 0;
}